/*-------------------------------------------------------------
 *  Copyright (c) 2011-2013, MSS ComBlock. All rights reserved.
 *
 * 	Filename: RF.c
 *  Last modified: 3/9/13
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 2
 *  Inheritance: N/A
 *  Description: RF components driver
 *
 *  Rev 1.1 6/1/11 AZ
 *  Moved RF gain control 2 from PWM 1.2 to PWM 1.4 (reason: hardware version change from prototype to production)
 *
 *  Rev 2 4/25/12 AZ
 *  Corrected bug in program_tuner_pll
 *
 *  Rev 3 3/9/13 AZ
 *  Read mixer power detector, even when AGC loop is open
 *  reduced AGC threshold to prevent saturation (wideband noise + signal)
 *  Note: make sure shield is soldered on all 4 edges to prevent oscillations.
 -------------------------------------------------------------*/
#include "main.h"
#include "rf.h"
#include "fpga.h"

inline int pll_strobe_check(void) {
	return FIO0PIN & 0x00040000; // pin P0[18]
}

long synth_check_lock(void) {
	return FIO1PIN & 0x20000000; // pin P1[29] ADF4350 synthesizer lock status
}

unsigned long get_freq(long index) {
	// reconstruct the RF frequency synthesizer frequency from the control registers.
	// index range 0 - 7
	// Returns the frequency expressed in Hz.
	long reg_no;
	unsigned long freq;
	long i;

	// points to control register with MSB
	if (index == 0)
		reg_no = 3; // frequency in REG3-0
	else if (index < 8)
		reg_no = 6 + 4 * index;
	else
		return 0; // invalid index

	for (i = 0; i < 4; i++) {
		freq <<= 8;
		freq &= 0xFFFFFF00;
		freq |= ((unsigned long) regv[reg_no--]) & 0x000000FF;
	}
	return freq;
}

void adf4350_reg(long regval) {
	// write 32 serial bits to the ADF4350 Integrated synthesizer and VCO

	int i;
	PIN_SYNTH_CLK_L(); // always start transaction with clock low
	PIN_SYNTH_LE_L(); // Latch enable - low  (initialized high at power up)

	for (i = 0; i < 32; i++) {
		if (regval & 0x80000000) {
			PIN_SYNTH_DATA_H(); //data high
			PIN_SYNTH_CLK_H(); //clock high
		} else {
			PIN_SYNTH_DATA_L(); //data low
			PIN_SYNTH_CLK_H(); //clock high
		}

		regval <<= 1;
		PIN_SYNTH_CLK_L(); //clock low
	}

	PIN_SYNTH_LE_H(); // Latch disable - high
}

void init_tuner_pll(void) {
	// one-time ADF4350 initialization.
	// 20 MHz frequency reference
	unsigned long regval;

	// program registers in sequence 5,4,3,2,1,0
	// register R5
	regval = 0x00400005;
	// LD pin = digital lock detect
	adf4350_reg(regval);

	// register R3
	regval = 0x00008013;
	// clock divider value = 400uS * 5 KHz = 2
	// fast clock enable
	// cycle slip reduction disabled (seems to increase phase noise a lot! when enabled)
	adf4350_reg(regval);

	// register R2
	// double 10 MHz ref clock for better phase noise)
	regval = 0x1B006E42; //  preference: low noise mode
	//regval = 0x7A006E42;   // preference: low spur mode
	// counter reset disabled
	// cp three-state disabled
	// not powered down
	// pd positive polarity, 10ns LDP
	// frac-N, icp = 2.5ma
	// double buffer enabled
	// 10-bit R counter = 1, ref doubling (for better phase noise) followed by 1:2
	// muxout = digital lock detect
	// low phase noise mode
	adf4350_reg(regval);

	// register R1
	regval = 0x0800FD01; // 8/9 prescaler, MOD = 4000
	adf4350_reg(regval);

	return;
}

void program_tuner_pll(void) {
	// program RF synthesizer with default frequency
	// One must call init_tuner_pll() once prior to using this function (one-time).

	unsigned long regval, f, fvcodiv2, fi;
	unsigned short nint, nfrac;

	// default frequency index is in regv[6] bits 7:5
	fi = (regv[6] & 0x000000E0) >> 5;	// corrected 4/25/12 AZ
	// construct frequency word (in Hz) from control registers
	f = get_freq(fi); // frequency in Hz
	// estimated receiver center frequency (needed to compute attenuators response)
	fc = f - 125000000;

	// program registers in sequence 5,4,3,2,1,0

	// register R4
	// convert tuner_freq into parameters understandable by the ADF4350
	// multiply turn tuner_freq to get a VCO frequency before the final output divider
	// Disable auxilliary output (RF_OUTB unused)
	// Set output LO power to -1dBm (trying to minimize LO leakage to the RF input).
	if (f >= 2200000000) {
		regval = 0x008C842C; // fout >= 2.2 GHz
		fvcodiv2 = f >> 1; // VCO frequency/2 (to prevent overflow later in the computation)
	} else if (f >= 1100000000) {
		regval = 0x009C842C; // fout >= 1.1 GHz
		fvcodiv2 = f;
	} else if (f >= 550000000) {
		regval = 0x00AC842C; // fout >= 550 MHz
		fvcodiv2 = f << 1;
	} else if (f >= 275000000) {
		regval = 0x00BC842C; // fout >= 275 MHz
		fvcodiv2 = f << 2;
	} else {
		regval = 0x00CC842C; // fout >= 137.5 MHz
		fvcodiv2 = f << 3; // VCO frequency/2 (to prevent overflow here when multiplying by 16)
	}
	// feedback select = fundamental
	// band select clock divider = 10MHz/125KHz max = 80. use 100.
	// vco powered up, mute till lock detect,
	// output power +5 dBm (both outputs)
	// aux output enabled, divided output
	adf4350_reg(regval);

	// register R0
	// reference frequency is 20 MHz (after doubling)
	nint = (unsigned short) (fvcodiv2 / 10000000);
	fvcodiv2 -= nint * 10000000;
	// frac step size is 5KHz
	nfrac = (unsigned short) (fvcodiv2 * 2 / 5000);
	regval = (unsigned long) (nint);
	regval <<= 12;
	regval = regval & 0x0FFFF000;
	regval |= ((unsigned long) (nfrac)) & 0x000FFF;
	regval <<= 3;
	regval &= 0x7FFFFFF8;
	adf4350_reg(regval);
}

void RF_synthesizer_output_enable(long en) {
	// enable/disable RF synthesizer output
	if (en) {
		PIN_SYNTH_ENABLE_H();
	} else {
		PIN_SYNTH_ENABLE_L();
	}
}

// set the frequency translation in the FPGA
// 32-bit signed value stored in reg36-39 + 4*i, where index i = 0 - 7
// Write to FPGA registers 100-103
void digital_frequency_translation(void) {
	unsigned long fi;

	// default frequency index is in regv[6] bits 7:5
	fi = (regv[6] & 0x000000E0) >> 5;	// corrected 4/25/12 AZ

	fpga_setByte(100, regv[36 + 4 * fi]);
	fpga_setByte(101, regv[37 + 4 * fi]);
	fpga_setByte(102, regv[38 + 4 * fi]);
	fpga_setByte(103, regv[39 + 4 * fi]);
}

void DAC_init(void) {
	// initialize internal 10-bit DAC
	long a;

	// 1. to enable the DAC, output must be selected to appear on the related pin p0.26
	PINSEL1 = (PINSEL1 & 0xFFCFFFFF) | 0x00200000;

	// 2. peripheral clock selection for DAC.
	// select the processor clock CCLK/4 (default)
	PCLKSEL1 = PCLKSEL1 & 0xFF3FFFFF;

	// 3. enable double buffering, no interrupts, 1us refresh
	DACCTRL = 0x00000006;
	DACCNTVAL = 30; // 1us @120MHz/4
}

void DAC(long value) {
	// output is value * Vrefp / 1024
	// 1us settling time -> 1 MHz maximum update rate
	DACR = (value << 6) & 0x0000FFC0;
}

void PWM_init(void) {
	// initialized P1[23] = PWM 1.4 as a fake 10-bit DAC (externally filtered)
	// PWM1IR interrupt register ignored (no interrupts)
	PCONP |= (1 << 6); // PWM1 power ON
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	PWM1TCR = 0x00000002; // Reset counter
	PWM1PR = 0;
	PWM1PC = 0;
	PINSEL3 = (PINSEL3 & 0xFFFF3FFF) | 0x00008000;
	// p1[23] = PWM1.4
	PINMODE3 = (PINMODE3 & 0xFFFF3FFF) | 0x00008000;
	// P1[23] PWM has no PU/PD
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR4 = 0; // initial PWM value
	// PWM Match Control Register
	// Match 0 is configured to reset the timer/counter when a match event occurs.
	PWM1MCR = 0x00000002; // reset counter when match MR0
	// PWM Control Register
	PWM1PCR = 0x00001000; // The PWM4 output enabled
	// PWM Timer Control Register
	PWM1TCR = 0x00000009; // clear counter,  enable PWM mode, enable counting
}

void PWM(long value) {
	// output is value * 3.3V / 1024 after filtering
	PWM1MR0 = 1024; // controls the PWM cycle rate. Here 1024 peripheral clocks
	PWM1MR4 = value & 0x000003FF; // controls the PWM edge position

	// PWM Latch Enable Register
	PWM1LER = 0x00000011; // Enable PWM Match 0 and 4 Latch
}

// Found a balance of 0x40/0xFF the best to prevent saturation (and the resulting harmonics)
// at lower frequencies
void setRFGain(void) {
	// breaks down the overall receiver gain into RF gain (two pin-diode attenuators)
	// and one amplitude gain control in the FPGA (see setBBgain)
	long gain_att1, gain_att2;

	// 1st attenuator
//	gain_att1 = RF_gain * 0x0040 / 256;
	gain_att1 = RF_gain >> 2;	// rephrase, simpler
	if (gain_att1 > 1023)
		gain_att1 = 1023;
	// 2nd attenuator
//	gain_att2 = RF_gain * 0x00FF / 256;
	gain_att2 = RF_gain;	// rephrase, simpler
	if (gain_att2 > 1023)
		gain_att2 = 1023;

	DAC(gain_att1);
	PWM(gain_att2);

}

void setBBGain(void) {
	// breaks down the overall receiver gain into RF gain (two pin-diode attenuators)
	// and one amplitude gain control in the FPGA (see setBBgain)
	unsigned char bbgainLSB, bbgainMSB;

	// set digital gain to FPGA
	bbgainLSB = (unsigned char)(BB_gain & 0x00FF);
	bbgainMSB = (unsigned char)((BB_gain >> 8) & 0x00FF);
	fpga_setByte(70, bbgainLSB);	// digital amplitude gain LSB
	fpga_setByte(71, bbgainMSB);	// digital amplitude gain MSB

}

inline void RF_AGC_poll(void) {
	// 3 AGC modes as set in REG69(7:6)
	unsigned char agc_loop, a;
	agc_loop = (regv[69] & 0xC0) >> 6;

	// check level/power sensors for reporting status (even when AGC loop is open) 3/9/13
	ADC_RF_POWER_DET();
	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
		// RF AGC loop algorithm relying on RF power detection (prevent saturation at
		// RF mixer) and peak level after the A/D converter (prevent saturation at the A/D converter)
	} else if ((agc_loop == 1) || (agc_loop == 2)) {
		// RF AGC loop under this processor control
		// read peak detect status from FPGA (SREG1)
		// The peak amplitude is detected over a 256-sample window at the ADC output = 64 signal periods.
		a = fpga_getByte(1);
		// read RF power detect rf_power_det (done periodically in main loop)
		// new: threshold reduced to prevent saturation (wideband noise + signal)
		if ((rf_power_det > 0x0B40) || (a > 0x00C0)) {
			// getting close to saturation, either at the RF mixer or at the A/D converter. Decrease gain
			RF_gain_too_large_duration++;
			RF_gain_too_small_duration = 0; // reset the time elapsed since last peak detect > 0x80
			if (RF_gain_too_large_duration > 16) { // agc reaction time implied here
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_decrement();
			}
		} else {
			// slow pressure to increase gain.
			RF_gain_too_small_duration++; // increment the time elapsed since last saturation detect
			if (RF_gain_too_small_duration > 16) { // agc reaction time implied here
				// gain is consistently too small.
				RF_gain_increment();
				RF_gain_too_large_duration = 0; // reset counter
				RF_gain_too_small_duration = 0; // reset the time elapsed since last saturation detect
			}
		}
	}
}

inline void BB_AGC_poll(void) {
	// 3 AGC modes as set in REG69(7:6)
	unsigned char agc_loop, ai, aq;
	agc_loop = (regv[69] & 0xC0) >> 6;

	if (agc_loop == 0) {
		// open loop. Gain is fixed.
		return;
		// Baseband AGC loop algorithm relies on detecting saturation or near saturation at the
		// output samples.
	} else if (agc_loop == 1) {
		// Baseband AGC loop under this processor control
		// read output samples MSBs from FPGA (SREG2/SREG3)
		ai = fpga_getByte(2);
		aq = fpga_getByte(3);
		if((ai > 0x00F8) || (ai < 0x0008) || (aq > 0x00F8) || (aq < 0x0008)){
			// Near saturation detected. Reduce gain
			if(BB_gain > 16)
				BB_gain -= 16;
		} else {
			// slow pressure to increase the gain
			if(BB_gain < 65535)
				BB_gain++;
		}
		// set digital gain
		setBBGain();

	} else if (agc_loop == 2) {
		// external AGC loop
		// read ADC (converting the AGC_IN signal from the 40-pin connector to a 12-bit value)
		ADC_AGC_IN();
		// flip gain control value (ComBlocks AGC rule: 0 is for maximum gain)
		BB_gain = 65535 - (agc_in << 4);	// agc_in range 0 - 4095
		// set digital gain
		setBBGain();
	}
}

void RF_gain_increment(void) {
	if (RF_gain < 4095) {
		RF_gain++;
		setRFGain();
	} else {
		RF_gain = 4095; // enforce range 0 -  4095
	}
}

void RF_gain_decrement(void) {
	if (RF_gain > 0) {
		RF_gain--;
		setRFGain();
	} else {
		RF_gain = 0; // enforce range 0 -  4095
	}
}

//long pinDiodeAttenuatorCompensation(long gain){
//	// attempt at avoiding instabilities
//	// function of center frequency fc (global)
//	// input range 0 - 1023
//
//	long revised_gain;
//	revised_gain = gain;
//
//	if(fc < 200000000){
//		// below 200 MHz, avoid very large attenuations (Vc < 0.55V) which are destabilizing
//		if(gain < 129)
//			revised_gain = 129;
//	} else if (fc < 500000000){
//		// below 500 MHz, avoid very large attenuations (Vc < 0.5V) which are destabilizing
//		if(gain < 108)
//			revised_gain = 108;
//	} else if (fc > 2000000000){
//		// above 2 GHz, avoid too much gain (> 2.5V) because it can create instabilities
//		if(gain > 538)
//			revised_gain = 538;
//	}
//	return revised_gain;
//}

void ADC_init(void) {
	// ADC initialization
	PCONP |= 0x00001000; // power ADC subsystem
	// PCLKSEL0: by default peripheral clock is CCLK/4 = 30 MHz
	// A/D Control Register: inputs are AD0[2] and [5], ADC clock is 10 MHz, burst=1, operational,
	AD0CR = 0x00210224;
}

void ADC_AGC_IN(void) {
	// returns the last value read on AD0[5] = AGC_IN from 40-pin connector, low-pass filtered
	unsigned long a;
	a = AD0DR5;
	if (a & 0x80000000) {
		// done bit high. update agc_in
		agc_in = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}

void ADC_RF_POWER_DET(void) {
	// returns the last value read on AD0[2] = RF power detected from the RF mixer
	unsigned long a;
	a = AD0DR2;
	if (a & 0x80000000) {
		// done bit high. update rf_power_det
		rf_power_det = (a >> 4) & 0x00000FFF; // 12-bit ADC measurement
	}
}

void setLPF() {
	// set the low-pass filter bandwidth
	// REG4/REG5 stores the one-sided bandwidth in kHz
	// Controls the CIC decimation filter and the two Half-Band filters
	int bw, cic_r;

	// read single-sided bandwidth in KHz
	bw = regv[5] & 0x00FF;
	bw = ((bw << 8) & 0x00FF00) | (regv[4] & 0x00FF);

	// assuming a 100 MSamples/s sampling clock at the FPGA, let's calculate the
	// decimation and the need to bypass
	// we want to decimate from +/-fs/2 (at the ADC) to +/-4*bw (at the CIC filter output)
	cic_r = (int) (50000 / (4 * bw));
	// check valid range 1 - 16384
	if (cic_r < 1)
		cic_r = 1; // 0 is not an acceptable value
	else if (cic_r > 16384)
		cic_r = 16384;
	regv[107] = (unsigned char) (cic_r & 0x00FF);
	regv[108] = (unsigned char) ((cic_r >> 8) & 0x003F);

	// bypass filters
	if (bw > 12500) {
		// bypass CIC + 1st HBF
		regv[108] = (regv[108] & 0xC0) | 0x40;
	} else if (bw > 6250) {
		// bypass CIC
		regv[108] = (regv[108] & 0xC0) | 0x80;
	} else {
		// enable all filters
		regv[108] = (regv[108] & 0xC0) | 0xC0;
	}

	fpga_setByte(107, regv[107]);
	fpga_setByte(108, regv[108]);

}
