/*-------------------------------------------------------------
 *  Copyright (c) 2010-2011, MSS ComBlock. All rights reserved.
 *
 * 	Filename: fpga.c
 *  Last modified: 3/2/11
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 1.0
 *  Inheritance: CB1600_2_00 1/10/11
 *  Description: Utilities to configure the FPGA.
 *  IMPORTANT ASSUMPTION: data file is BIT SWAPPED (select BIT SWAP when generating the bit/mcs file in Xilinx ISE)
 *
 *  Dependencies/Pre-requisite: N/A
 -------------------------------------------------------------*/
#include "main.h"
#include "fpga.h"

/*-------- change this section according to the schematics ----------------------------------------*/
inline long fpga_check_initb(void){
	return FIO0PIN & 0x00000080;	// pin P0[7] INIT#
}

inline long fpga_check_done(void){
	return FIO0PIN & 0x00000040;	// pin P0[6] DONE
}

inline void initb2web(void){
	// after configuration change init# input pin to we# output
	FIO0DIR |= 0x00000080;	// p0[7] is now an output
	PIN_FPGA_WEB_H();
	return;
}

// Select/Deselect FPGA communication
inline void FPGAsel(void){
	PIN_FPGA_CSIB_L();	// enable FPGA
}
inline void FPGAdeselect(void){
	PIN_FPGA_CSIB_H();	// disable FPGA
}

inline void busDirOut(void){
	// set 8-bit data bus direction to 'out'
	FIO2DIR |= 0x000003FF;	// UC_DATAx + UC_CTRL as outputs
}

inline void busDirIn(void){
	// set 8-bit data bus direction to 'in'
	FIO2DIR &= ~0x000000FF;	// UC_DATAx as inputs
}

inline unsigned char busDataIn(void){
	return (unsigned char)(FIO2PIN);
}

inline void busDataOut(unsigned char dout){
	FIO2PIN = (FIO2PIN & 0xFFFFFF00) | (((long)(dout)) & 0x000000FF);
}

/*-------- end of schematics-dependent code ----------------------------------------*/


long fpga_config(){
	// configure the FPGA with the configuration data stored in internal LPC1759 flash memory
	unsigned long rptr, start_addr, end_addr;
	unsigned long i;
	unsigned char pageBuf[1024];	// page from the flash memory

	// Flash start/end addresses
	start_addr = FPGA_START_ADDR;	// assumes that all ARM code fits below the FPGA configuration in flash memory
	end_addr = start_addr + FPGA_CONFIG_SIZE;
	if(end_addr > 0x0007FFFF)
		end_addr = 0x0007FFFF;	// internal flash limited to 512KB
	// the Xilinx bit file may actually be 30% or so smaller since compression is used.
	// See UG380 BitGen option -g compress. [ISE: in Generate Programming File | Process Properties ]
	// Compression MUST be used when generating the bit file since the flash memory is only 512KB for all uses.

	// The key waveforms are specified in Xilinx UG380 Figure 2-9

	// prepare FPGA for programming
	PIN_FPGA_PROGB_H();
	PIN_FPGA_RWB_H();
	PIN_FPGA_CSIB_H();
	PIN_FPGA_CCLK_H();

	// initialize FPGA
	PIN_FPGA_PROGB_L();
	PIN_FPGA_RWB_L();
	PIN_FPGA_CCLK_L();	// needed for sure (for timing?)
	PIN_FPGA_CCLK_H();
	PIN_FPGA_PROGB_H();

	// await INIT# going high (measured 600us)
	timer[0] = 2; 	// 10ms units -> 10-20ms timeout
	while(1){
		if(timer[0] == 0){	// timer expired
			return -FPGA_ETIMEOUT; // timeout
		}
		if(fpga_check_initb())
			break;
	}

	busDirOut();
	PIN_FPGA_RWB_L();	// make sure that RWB is always in the write position when CSIB is enabled (otherwise it means abort)
	PIN_FPGA_CSIB_L();	// enable FPGA

	// read one page at a time
	for(rptr = start_addr; rptr <= end_addr; rptr += sizeof(pageBuf)){
		// read a page from the internal flash memory
		memcpy(&pageBuf[0], (void *)(rptr), sizeof(pageBuf));
		//FIO1CLR = 0x00000002;	// TEST TEST TEST

		/* before timing optimization
		for(i = 0; i < sizeof(pageBuf); i++){
			busDataOut(pageBuf[i]);	// IMPORTANT ASSUMPTION: data file is BIT SWAPPED (select BIT SWAP when generating the bit/mcs file in Xilinx ISE)
			PIN_FPGA_CCLK_L();
			PIN_FPGA_CCLK_H();
		}
		*/
		/* after timing optimization */
		for(i = 0; i < sizeof(pageBuf); i++){
			FIO2PIN = pageBuf[i] & 0x000001FF;
			PIN_FPGA_CCLK_H();
		}
		/* end of timing optimization */

	}

	// disable all FPGA control lines
	PIN_FPGA_CSIB_H();
	PIN_FPGA_RWB_H();
	// read back (debug only, we need the SelectMAP pins for I/O during operations).
	// fpga_readback(); // untested
	initb2web();	// init# input is now we# output
	return FPGA_OK;
}


long fpga_readback(void){
	// readback FPGA configuration.
	// Three key requirements: BitGen security setting must not prohibit readback (-g Security:none), and bitstream
	//  encryption must not be used. Additionally, the port must be set to retain its function after configuration by
	// setting the persist option in BitGen (-g Persist:Yes), otherwise the SelectMAP data pins
	// revert to user I/O, precluding further configuration operations.
	// See Xilinx ISE generate programming file | properties GUI
	unsigned long i, rptr, start_addr, end_addr;
	unsigned char pageBuf[1024];	// page from the internal flash memory

	start_addr = FPGA_START_ADDR;	// assumes that all ARM code fits below the FPGA configuration in flash memory
	end_addr = start_addr + FPGA_CONFIG_SIZE;
	if(end_addr > 0x00080000)
		end_addr = 0x00080000;	// internal flash limited to 512KB

	// waveform as per ug380 Figure 6-1
	PIN_FPGA_CSIB_H();
	PIN_FPGA_RWB_H();
	PIN_FPGA_CCLK_L();
	PIN_FPGA_CCLK_H();
	busDirIn();
	// read one page at a time
	for(rptr = start_addr; rptr <= end_addr; rptr += sizeof(pageBuf)){
		for(i = 0; i<sizeof(pageBuf); i++){
			PIN_FPGA_CCLK_L();
			PIN_FPGA_CCLK_H();
			pageBuf[i] = busDataIn();
		}
	}
	PIN_FPGA_CSIB_H();
	return 0;
}

long block_to_config(unsigned long blockNo){
	// returns the configuration number (starting at 0) based on the flash block number
	// Only one configuration in COM-3011
	return 0;
}

void fpga_reset(void){
	// erase FPGA configuration
	PIN_FPGA_PROGB_L();
	PIN_FPGA_RWB_L();
	PIN_FPGA_CCLK_L();	// needed for sure (for timing?)
	PIN_FPGA_CCLK_H();
	PIN_FPGA_PROGB_H();
	return;
}

unsigned char fpga_getByte(unsigned long addr){
	// read byte from the specified FPGA address
	// The number of addresses supported depend on the FPGA code.
	unsigned char a;

	PIN_FPGA_REB_H();	// read#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable
	busDirOut();
	FPGAsel();	// enable FPGA

	busDataOut((unsigned char)(addr));
	PIN_FPGA_ALE_H();	// address latch enable
	PIN_FPGA_WEB_L();	// write#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable

	busDirIn();
	PIN_FPGA_REB_L();	// read#
	a = busDataIn();
	PIN_FPGA_REB_H();	// read#
	PIN_FPGA_CSIB_H();	// deselect FPGA
	return a;
}

void fpga_setByte(unsigned long addr, unsigned char a){
	// write a byte to the specified FPGA address
	// The number of addresses supported depend on the FPGA code.

	PIN_FPGA_REB_H();	// read#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable
	busDirOut();
	FPGAsel();	// enable FPGA

	busDataOut((unsigned char)(addr));
	PIN_FPGA_ALE_H();	// address latch enable
	PIN_FPGA_WEB_L();	// write#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_ALE_L();	// address latch enable

	busDataOut(a);
	PIN_FPGA_WEB_L();	// write#
	PIN_FPGA_WEB_H();	// write#
	PIN_FPGA_CSIB_H();	// deselect FPGA
	return;
}

long fpga_config_exists(unsigned long start_addr){
	// check the first few bytes in internal flash to verify that the FPGA configuration exists.
	// otherwise, there one could damage the FPGA by programming all FFs.
	unsigned char pageBuf[1024];	// page from the NAND memory
	long retval, i;

	memcpy(&pageBuf[0], (void *) (start_addr), sizeof(pageBuf));
	for(i = 0; i<sizeof(pageBuf); i++){
		if(pageBuf[i] != 0xFF)
			return 1;	// non-FF character present
	}
	return 0;	// blank page.
	// It is most likely that fpga configuration is not programmed into the NAND flash memory,
	// at least at this address.


}
