/*-------------------------------------------------------------
 *  Copyright (c) 2010-2016, MSS ComBlock. All rights reserved.
 *
 * 	Filename: main.c
 *  Last modified: 3/14/16
 *  Authors:
 *  	Alain Zarembowitch / MSS ComBlock
 *  Version: 3.01
 *  Inheritance: comblock1400 Atmel code
 *  Description: main program. Includes main loop for handling commands on a polling basis.
 *
 *  Dependencies/Pre-requisite: N/A
 *
 *  Rev 2 4-24-11 AZ
 *  Editorial changes to minimize the number of warnings.
 *  Implemented independent RF and Baseband AGC loops.
 *
 *  Rev 1_04  5/12/12 AZ
 *  Bug correction: trigger the FPGA initial configuration by writing to control reg 71
 *  Pull-down on VBUS (otherwise too slow to discharge when USB cable is removed)
 *
 *  Rev 2.03 12/12/15 AZ
 *  Added usb connection/disconnection in main, depending on the presence of the USB VBUS
 *
 *  Rev 3.01 3/14/16 AZ
 *  Added forgotten feature (PLL_STROBE)
 -------------------------------------------------------------*/
#define __MAIN_C__

#include "main.h"
#include "uart.h"
#include "utils.h"
#include "iap.h"
#include "fpga.h"
#include "rf.h"

/*=========================================================================*/
/*  DEFINE: All Structures and Common Constants                            */
/*=========================================================================*/
volatile unsigned long SysTickCnt; /* SysTick Counter                    */
unsigned long cpuFreq; // CPU frequency in Hz.
volatile unsigned long timer[NTIMERS];
// timers, expressed in multiple of the system tick (10ms). 0 means expired.
// volatile because the Systick ISR will modify the values
// timer[0] = flash timer
// timer[1] = forced delay to save the config into flash
// timer[2] = inactivity timer
volatile string_struct rxLine; // rx command line
volatile string_struct txLine; // tx command line
// separate rx and tx lines (unlike previous Atmel code). Allows concurrent tx and rx and interrupts.
char address[4]; // ComBlock address, assigned dynamically (string of 3 characters + \0 end of string)
long commandPortNo; // port through which this board receives commands: 1 = top (USB ARM), 2 = right, 3 = bottom (USB FPGA), 4 = left (ComBlock rule)
unsigned char forwardMask; // Mask output ports. bit0 = top port1, bit1 = right port 2, etc. '1' means the port is blocked, '0' means the port is enabled for output
struct config_hw configHw; // hardware-related configuration.
struct config_sw configSw; // software/FPGA-related configuration volatile
char regv[NREGS]; // control registers, volatile copy of configSw.reg[]
int active_sw_config; // current active configuration (generally the same as default_sw_config, but it could be temporary different until the new FPGA config is loaded)
int usb_softconnect;	// 0 for no vbus/no connect, 1 for vbus/connected
unsigned long RF_gain;	// receiver RF gain (2 RF attenuators + 1 digital gain control)
unsigned long BB_gain;	// receiver baseband gain (1 digital gain control)
long RF_gain_too_small_duration; // below normal range samples timer
long RF_gain_too_large_duration; // above normal range samples timer
long fc;	// estimated receiver center frequency (needed to compute attenuators response)
unsigned long agc_in;	// gain control value from the 40-pin connector, low-pass filtered = AD0[5]
unsigned long rf_power_det;	// rf power detection from RF mixer = AD0[2]

/***************************************************************************/
/*  main                                                                   */
/***************************************************************************/
int main(void) {
	// local variables
	unsigned char a;
	long retval;
	unsigned int i, j, c, reg_no, s78_decode, s789_decode, packet_no,
			rx_byte_count, l, sect_no, pll_strobe_last_val, freq_index;
	int sn[4];
	int sel;
	struct config_sw cs; // work area for a software configuration (don't store anything of value here)
	char buf1[128]; // temporary storage for WFS command
	char sreg2, sreg4, sreg6;	// temporary storage for multi-byte monitoring values

	// initial ARM processor configuration (clock frequency, system tick, etc)
	initConfig();
	// reset FPGA configuration (just in case it was interfering with the ARM - NAND Flash shared data bus
	rst: fpga_reset();


	/* POSSIBLE CHANGES NEXT REV
	 *
	 * 2. double check if USB_CONNECT with external MOSFET is enabled. If not, copy COM-3011 code.
	 *
	 */






	// initialize UARTs 0,2,3. UART2 is reserved for future use (demodulated data in a self-standing receiver for example).
	uart_init();

	// When the internal flash sector CONFIG_SECTOR is blank (all FFs), initialize it to 'all zeros'.
	initialize_configs();
	// load last-saved configuration from non-volatile (flash) memory
	read_hw_config(&configHw);
	// assess JP1 once at power-up/reset. Use get_JP1val() to read jumper location.
	assess_JP1();

	// Only one personality in the COM-3011.
	active_sw_config = 0;

	// Semantics: we make a difference between the active_sw_config (currently loaded and running) and
	// default_sw_config which can be changed by the user prior to FPGA reboot. N/A here (one configuration)
	read_sw_config(active_sw_config, &configSw);
	memcpy(&regv, configSw.reg, sizeof(regv)); // initialize the temporary control registers with the previously saved ones.

	// configure the FPGA
	// skip the initial FPGA configuration if JP1 is set to position 1-2 or if the software configuration is undefined
	if (get_JP1val() != 12) {
		// configure FPGA
		fpga_config();
	}

	// add a small delay (> 10 ms) to make sure the FPGA PLL is up and running.
	delay_ms(50);

	// ComBlock address at power up is "000"
	snprintf(&address[0],sizeof(address), "000");

	// configure the FPGA with the 250 registers stored in non-volatile memory
	// skip registers 250,251,....,254 & 255, they are used to rx/tx data from FPGA
	for (i = 0; i < 100; i++) {
		fpga_setByte(i, configSw.reg[i]);
	}
	for (i = 200; i < 250; i++) {
		fpga_setByte(i, configSw.reg[i]);
	}

	// RF frequency synthesizer (ADF4350) initialization
	init_tuner_pll(); // one time initialization
	program_tuner_pll(); // Program the default frequency
	RF_synthesizer_output_enable(1);	// enable the output

	// digital frequency translation
	digital_frequency_translation();

	// set CIC decimation and HBF
	setLPF();

	// RF gain control. DAC and PWM initialization. Use REG[68] and REG[69] for initial setting
	DAC_init();
	PWM_init();
	RF_gain = (long)(regv[69]) & 0x0000000F;	// initial gain
	RF_gain <<= 8;
	RF_gain |= (long)(regv[68]) & 0x000000FF;
	setRFGain();	// set initial RF gain

	// Baseband gain control. Use REG[70] and REG[71] for initial setting
	BB_gain = (long)(regv[71]) & 0x000000FF;	// initial gain
	BB_gain <<= 8;
	BB_gain |= (long)(regv[70]) & 0x000000FF;
	setBBGain();	// set initial baseband gain

	// ADC initialization
	ADC_init();

	// write (again) to the last control register in the FPGA to activate the configuration
	fpga_setByte(71, regv[71]);

	// main loop -------------------------------------------------------------------------------
	while (1) {
		// USB softconnect management based on USB VBUS presence
		USB_connect_ctrl();

		// is it time to save the current active configuration?
		// timer[1] limits how frequently the configuration can be saved,
		// timer[2] indicates persistent inactivity.
		if (((configHw.save_request_pending) || (configSw.save_request_pending))
				&& (timer[2] == 0) && (timer[1] == 0)) {
			save_configuration(-1, &cs); // save the hardware + active software configurations. no alternate configuration.
		}

		// service UARTs (polling)
		uart_rxpoll();
		uart_txpoll();

		// RF gain control. Read power sensors AD0[2] and AD0[5]
		RF_AGC_poll();
		// Baseband gain control
		BB_AGC_poll();

		// pll strobe (polling)
		if((regv[6] & 0x04) == 0x04){
			// external control enabled
			if((pll_strobe_check() != 0) && (pll_strobe_last_val == 0)) {
				// detected rising edge of PLL_STROBE signal
				freq_index = (regv[6] >> 5) & 0x07;
				freq_index++;
				if(freq_index >= regv[35]){
					freq_index = 0;	// modulo REG35
				}
				regv[6] = (regv[6] & ~0x00E0) | ((freq_index << 5) & 0x00E0);
				// reprogram PLL
				program_tuner_pll();
			}
			pll_strobe_last_val = pll_strobe_check();
		}

		// did we receive any message?
		// get a complete message from @ to \n (non-blocking)
		if ((rxLine.msg_complete) && (rxLine.rptr == 0)) {
			// When done processing the received message in rxLine, move the rptr = wptr to allow subsequent messages to be received.

			// re-arm inactivity timer because of new message just received.
			// We need to know when there is inactivity so that we can devote 120ms to update the configuration in flash.
			timer[2] = TIMER2;

			/*
			 // echo message (debugging)
			 memcpy(txLine.buf, rxLine.buf, rxLine.wptr);
			 txLine.wptr = rxLine.wptr;
			 rxLine.rptr = rxLine.wptr;	// release rx message
			 sendLine(4);
			 continue;
			 */

			// Dispatch the message: local processing, ignore or forward?
			if (dispatchLine()) {
				// rx message is for us: address at the start of the rx message matches (exactly or 111 broadcast address)
				delay_ns(25000); // delay 25uS before responding (otherwise slower Atmel-based ComBlocks may not be ready for a reply)

				// Remember which port is used to communicate with ComBlock control center
				// Criteria: received complete command message destined to this particular ComBlock.
				commandPortNo = rxLine.portNo;

				// decode 2-digit decimal reg_no at position 7-8 in the received string. Used by most commands.
				s78_decode = rxLine.buf[7] - '0';
				s78_decode = s78_decode * 10 + rxLine.buf[8] - '0';

				// decode 3-digit decimal sector number at position 7-9 in the received string (common to many commands)
				s789_decode = s78_decode * 10 + rxLine.buf[9] - '0';
				sect_no = s789_decode + 16; // translate ComBlock range 001-999 to internal flash block number range starting at 17.


				// command handler
				if (!memcmp((const char *)&rxLine.buf[4], "SAC", 3)) {
					// SAC command - set address
					memcpy(&address[0], (const char *)&rxLine.buf[7], 3);

				} else if (!memcmp((const char *)&rxLine.buf[4], "MFW", 3)) {
					// MFW command - Message Forwarding
					// 0 = no forwarding, 1-4: forward to port 1-4
					// 9: forward to all ports (except of course the one where the message came from)
					if (rxLine.buf[7] == '0') { // stop all forwarding
						forwardMask = 0xFF;
					} else if (rxLine.buf[7] == '1') { // forward to top port
						forwardMask = 0xFE;
					} else if (rxLine.buf[7] == '2') { // forward to right port
						forwardMask = 0xFD;
					} else if (rxLine.buf[7] == '3') { // forward to bottom port
						forwardMask = 0xFB;
					} else if (rxLine.buf[7] == '4') { // forward to left port
						forwardMask = 0xF7;
					} else if (rxLine.buf[7] == '9') { // forward to all ports, but the command port itself
						if (commandPortNo == 1) {
							forwardMask = 0xF1;
						} else if (commandPortNo == 2) {
							forwardMask = 0xF2;
						} else if (commandPortNo == 3) {
							forwardMask = 0xF4;
						} else if (commandPortNo == 4) {
							forwardMask = 0xF8;
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "PNG", 3)) {
					// new PNG ping command. Echoes command back
					memcpy((char *)&(txLine.buf[0]), (const char *)&(rxLine.buf[0]), rxLine.wptr);
					txLine.buf[1] = '9';
					txLine.buf[2] = '9';
					txLine.buf[3] = '9';
					txLine.wptr = rxLine.wptr;
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GMI", 3)) {
					// GMI - Get Module Identification command
					// Three cases:
					// (a) if the GMI command is followed by a non-zero FPGA configuration number,
					// then return the available information about that configuration,
					// (b) if the GMI command is followed by a zero configuration number,
					// then return the available information about the hardware module (FPGA platform)
					// (c) otherwise [if the GMI command is followed by CR or LF]
					// return the information about the active configuration
					// compute configuration number. comprises at exactly two digits.
					sel = 0;
					if ((rxLine.buf[7] != '\r') && (rxLine.buf[7] != '\n')) {
						sel = s78_decode;
					} else {
						// no configuration selected, asking about active configuration
						sel = -1;
					}
					if (sel == 0) {
						// user is asking about hardware platform description
						// This is a COM-1X00 hardware platform with ARM rev x
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MID" ARM_FW_REV "\r\n");
						txLine.wptr = strlen((const char *)txLine.buf); // set message length
						sendLine(commandPortNo); // send response through the same port the query came in
					} else if ((sel <= NSWCONFIGS) && (sel >= -1)) {
						// user is asking about a software configuration
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MID");
						if (sel == -1) {
							// no configuration selected, asking about active configuration
							memcpy(&cs, &configSw, sizeof(configSw));
							memcpy((char *)&txLine.buf[7], (const char *)&cs.mid[0], 4); // read 4-character mid
							txLine.buf[11] = fpga_getByte(0xfc); // option is read directly from FPGA itself
							txLine.buf[12] = fpga_getByte(0xfd); // revision is read directly from FPGA itself
						} else {
							// asking about a specific (not necessarily active) configuration
							sel--; // indices start at zero
							read_sw_config(sel, &cs); // read the software configuration of interest from flash
							memcpy((char *)&txLine.buf[7], (const char *)&cs.mid[0], 5); // read 5-character mid
							// non active configuration. We don't know the firmware rev no -> mask it
							txLine.buf[12] = ' ';
						}
						txLine.buf[13] = '0' + (unsigned char) (commandPortNo);
						// read configuration status: authorization/protection/etc  .
						// convert from binary to readable hex character
						// bit 0 = authorization
						// bit 1 = protection
						// bit 2 = FPGA DONE pin
						if (fpga_check_done())
							a = 0x04; // FPGA DONE pin is high
						else
							a = 0x00; // FPGA DONE pin is low

						// configuration is protected if boot section is protected (PROTECTION == 1)
						// or if protection flag is set for this configuration.
						if ((sel == 0) && (PROTECTION == 1)) {
							a = a | 0x02; // sector is protected
						} else if ((cs.protection) && (cs.mid[0] != 0xFF)) {
							a = a | 0x02; // module identification is defined and sector is protected
						}

						// TEMPORARY: authorized until hash code is in place
						a = a | 0x01; // always authorized (temporary)
						snprintf((char *)&txLine.buf[14], sizeof(txLine.buf) - 14, " %1X\r\n", a);
						// defensive code.. make sure there are no 0s. This could happen if the FPGA is not configured, or..
						for (i = 7; i <= 12; i++) {
							if (txLine.buf[i] == 0)
								txLine.buf[i] = '?';
						}
						txLine.wptr = strlen((const char *)txLine.buf); // set message length
						sendLine(commandPortNo);
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SDC", 3)) {
					// SDC command - set default configuration
					// decode selected Configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if ((c != configHw.default_sw_config) && (c < NSWCONFIGS)) {
						// the default config changed.
						configHw.default_sw_config = c;
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						// immediate save (as this command is likely to be followed by an immediate reset)
						timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
						// Note: saving takes 120ms!! during which incoming messages are not processed
						save_configuration(-1, &cs);
						//NOTE: at this point, active and default configuration numbers are different.
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GDC", 3)) {
					// GDC command - Get Default Configuration command
					// prepare DCF response
					c = configHw.default_sw_config + 1; // display configuration number starting at 1 (more natural for the user)
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999DCF%02X\r\n", c);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "SMI", 3)) {
					// SMI command - Set Module Identification

					// first check the length of the message.
					if ((rxLine.buf[12] == '\r') || (rxLine.buf[12] == '\n')) {
						// message is of the form @xxxSMI + MID + 1character option
						// thus 12 or 13 characters. use active configuration.
						// verify that the selected configuration is not protected unless
						// it is the first time this configuration is defined (all FF's)
						if ((!configSw.protection) || (configSw.mid[0] == 0xFF)) {
							memcpy(&configSw.mid[0], (const char *)&rxLine.buf[7], 5); // store MID + 1character option
							if (rxLine.buf[11] == '\r') // short message @xxxSMIYYYY\r\n
								configSw.mid[4] = ' '; // option cannot be '\r'
							configSw.mid[5] = 0; // don't know about FPGA firmware revision here
							configSw.protection = 0; // unprotected if first-time definition
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					} else if ((rxLine.buf[14] == '\r') || (rxLine.buf[14]
							== '\n')) {
						// message is of the form @xxxSMI+configuration number + MID + 1character option
						// thus 15 characters
						c = s78_decode - 1; // array indices start at 0, whereas configurations start at 1
						// make sure configuration pointer does not exceed array boundaries
						if (c < NSWCONFIGS) {
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							// verify that the selected configuration is not protected unless it is the first time this configuration is defined (all FF's)
							if ((!cs.protection) || (cs.mid[0] == 0xFF)) {
								memcpy(&cs.mid[0], (const char *)&rxLine.buf[9], 5); // store MID + 1character option
								cs.mid[5] = 0; // don't know about FPGA firmware revision here
								cs.protection = 0; // unprotected if first-time definition
								// immediate save this non-active configuration
								timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
								// Note: saving takes 120ms!! during which incoming messages are not processed
								save_configuration(c, &cs);
							}
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAI", 3)) {
					// SAI command - Set assembly identification
					// For security, allow only if not already programmed
					// first check if assembly id has already been programmed
					a = 0;
					/* temporarily disabled
					 for(i = 0; i < 4; i++){
					 if(aid[i] != 0)
					 a = 1;
					 }
					 */
					if (a == 0) {
						// if aid is undefined (all 0s), allow programming
						memcpy(&configHw.aid[0], (const char *)&rxLine.buf[7], 6);
						configHw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
						timer[1] = TIMER1SLOW;
						// re-arm timer[1]: wait before saving the comblock configuration
						//into flash (to consolidate multiple successive changes into one sector programming operation)
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "SAS", 3)) {
					// SAS - Set Assembly Serial number,
					// For security, allow only if eeprom is not already programmed (i.e. 0000..)
					// first check if serial number has already been programmed
					a = 0;
					for (i = 0; i < sizeof(configHw.asn); i++) {
						if (configHw.asn[i] != 0)
							a = 1;
					}
					if (a == 0) {
						// if assembly serial number is null, allow programming
						memcpy(&configHw.asn[0], (const char *)&rxLine.buf[7],
								sizeof(configHw.asn));
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAI", 3)) {
					// GAI command - Get Assembly ID
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999AID      \r\n"); // workaround problem with sprintf.
					memcpy((char *)&txLine.buf[7], (const char *)configHw.aid, 6);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GSN", 3)) {
					// GSN command - Get Serial Number
					// read unique LPC17xx device serial number
					read_sn(sn);
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999MSN%10u %10u %10u %10u\r\n",
							sn[0], sn[1], sn[2], sn[3]); // unsigned decimal
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GAS", 3)) {
					// GAS - Get Assembly Serial number command
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ASN%10s\r\n", configHw.asn);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if ((!memcmp((const char *)&rxLine.buf[4], "SRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "SRT", 3))) {
					// SRG - Set Register (persistent/non-volatile storage)
					// SRT - Set Register Temporary (i.e. volatile, lost when powering off)

					// compute register address. Address comprises at least two digits.
					c = s78_decode;
					// check whether destination register is 2 digit or 3 digit long
					if ((rxLine.buf[11] != '\r') && (rxLine.buf[11] != '\n')) {
						// three-digit register number
						c = c * 10 + rxLine.buf[9] - '0';
						// compute register value. Value starts at char 10
						a = hex2bin((char *)&rxLine.buf[10]);
					} else {
						// two-digit register number
						// compute register value. Value starts at char 9
						a = hex2bin((char *)&rxLine.buf[9]);
					}
					reg_no = c;
					// update the configuration register
					if (reg_no < sizeof(configSw.reg)) {
						regv[reg_no] = a;
						// write value to FPGA
						fpga_setByte(reg_no, a);
						// if SRG, also save in flash
						if (rxLine.buf[6] == 'G') {
							configSw.reg[reg_no] = a;
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						}
					}

					// REG6
					if(reg_no == 6){
						RF_synthesizer_output_enable((~a) & 0x01);	// enable or disable the RF synthesizer output
					}

					// writing to the last register of parameter triggers respective programming
					if (reg_no == 71) {
						program_tuner_pll(); // Program and start the pll
						digital_frequency_translation(); 	// NCO frequency in FPGA

						// set CIC decimation and HBF
						setLPF();

						// receiver RF gain. Initial value before RF AGC takes over. range 0 - 4095
						RF_gain = (long)(regv[69]) & 0x0000000F;
						RF_gain <<= 8;
						RF_gain |= (long)(regv[68]) & 0x000000FF;
						setRFGain();	// set initial RF gain

						// receiver baseband gain. Initial value before baseband AGC takes over. range 0 - 65535
						// Baseband gain control. Use REG[70] and REG[71] for initial setting
						BB_gain = (long)(regv[71]) & 0x000000FF;	// initial gain
						BB_gain <<= 8;
						BB_gain |= (long)(regv[70]) & 0x000000FF;
						setBBGain();	// set initial baseband gain

					}


				} else if (!memcmp((const char *)&rxLine.buf[4], "GSR", 3)) {
					// GSR command - Get Status Register
					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					// alternative status registers used for production tests (unused afterwards)
					a = 0;
					if (reg_no == 0) {
						// RF synthesizer lock status, FPGA configured, power good signals
						if (synth_check_lock())
							a |= 0x01; // SREG0 bit 0
						if (fpga_check_done())
							a |= 0x02; // SREG0 bit 1
						// power check
						a |= power_check();
					} else	if (reg_no == 1) {
						// receiver RF gain
						// save MSbs of RF_gain for SREG2
						sreg2 = (RF_gain >> 8) & 0x0F;
						a = RF_gain & 0x00FF;
					} else	if (reg_no == 2) {
						a = sreg2;
					} else	if (reg_no == 3) {
						// receiver baseband gain
						// save MSbs of BB_gain for SREG4
						sreg4 = (BB_gain >> 8) & 0x00FF;
						a = BB_gain & 0x00FF;
//					} else	if (reg_no == 3) {
//						// AGC_IN signal from the 40-pin connector, low-pass filtered
//						// save MSbs for later
//						sreg4 = (agc_in >> 8) & 0x0F;
//						a = agc_in & 0x00FF;
					} else	if (reg_no == 4) {
						a = sreg4;
					} else	if (reg_no == 5) {
						// RF power detect from the RF mixer
						// save MSbs for later
						sreg6 = (rf_power_det >> 8) & 0x0F;
						a = rf_power_det & 0x00FF;
					} else	if (reg_no == 6) {
						a = sreg6;
					} else {
						a = fpga_getByte(reg_no);
					}
					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);


				} else if ((!memcmp((const char *)&rxLine.buf[4], "GRG", 3)) || (!memcmp(
						(const char *)&rxLine.buf[4], "GRT", 3))) {
					// GRG command - get register (read from flash memory)
					// GRGT command - get register temporary (read from volatile memory)

					// check whether 1,2 or 3 digits
					c = rxLine.buf[7] - '0';
					if ((rxLine.buf[8] != '\r') && (rxLine.buf[8] != '\n')) {
						// 2 or 3 digit register number
						c = c * 10 + rxLine.buf[8] - '0';
					}
					if ((rxLine.buf[9] != '\r') && (rxLine.buf[9] != '\n')) {
						// 3 digit register number
						c = c * 10 + rxLine.buf[9] - '0';
					}
					reg_no = c;

					if ((reg_no < sizeof(configSw.reg)) && (rxLine.buf[6]
							== 'G'))
						a = configSw.reg[reg_no]; // GRG, read non-volatile register
					else if ((reg_no < sizeof(regv)) && (rxLine.buf[6] == 'T')) // make sure we do not write past the buffer end
						a = regv[reg_no]; // GRT, read volatile register
					else
						// trying to read non-volatile registers past index NREG_V
						a = 0;

					snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999RGV%02X\r\n", a);
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "EFS", 3)) {
					// EFS command - Erase File Sector
					// first check block number is valid
					// flash sectors 17 through 29 (32KB*13 = 416KB) are reserved for one FPGA configuration
					// Important: use the compress feature in Xilinx ISE to reduce file size.
					if ((sect_no < 17) || (sect_no > 29)) {
						retval = -1; // NAK
					} else {
						// make sure the block does not relate to a protected configuration
						c = block_to_config(sect_no);
						read_sw_config(c, &cs); // read the software configuration of interest from flash
						if ((c == 0) && (PROTECTION == 1)) {
							retval = -1; // boot configuration is protected
						} else if (cs.protection) {
							retval = -1; // configuration is protected
						} else {
							// IAP Erase Sector(s) command
							retval = internalFlash_EraseSector(sect_no);
						}
					}
					if (retval) {
						// error -> NAK
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ACK\r\n");
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "RMD", 3)) {
					// RMD - Read Memory Data command
					// write memory offset to FPGA registers 237 (LSB),238,239 (MSB)
					a = hex2bin((char *)&rxLine.buf[13]);
					fpga_setByte(237, a); // write value to FPGA  REG237 LSB
					a = hex2bin((char *)&rxLine.buf[11]);
					fpga_setByte(238, a); // write value to FPGA  REG238
					a = hex2bin((char *)&rxLine.buf[9]);
					fpga_setByte(239, a); // write value to FPGA  REG239
					// byte in input_string[7:8] ignored. too large an MSB. Not supported in this FPGA

					// compute length of sequence to read
					// ignore 2 MSBs (not supported, as FPGA has only a few 4Kb dual-port RAM buffers
					l = 0;
					l |= hex2bin((char *)&rxLine.buf[19]) & 0x000000FF;
					l <<= 8;
					l |= hex2bin((char *)&rxLine.buf[21]) & 0x000000FF;

					// check length
					if (l > 128) {
						// return message would be too long
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						snprintf((char *)&txLine.buf[0], sizeof(txLine.buf), "@999MDA");
						for (i = 0; i < l; i++) {
							// read REG250 for data from FPGA DPRAM
							// autoincrement on
							a = fpga_getByte(0xFA);
							snprintf((char *)&txLine.buf[7 + 2 * i], 5, "%02X\r\n", a);
						}
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "SCA", 3)) {
					// SCA - Set Configuration start Address command
					// read the selected configuration number for the specified start address
					// decode selected configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if (c < NSWCONFIGS) {
						// decode 32-bit hex string, then save
						l = 0;
						l |= hex2bin((char *)&rxLine.buf[9]) & 0x000000FF;
						l <<= 8;
						l |= hex2bin((char *)&rxLine.buf[11]) & 0x000000FF;
						l <<= 8;
						l |= hex2bin((char *)&rxLine.buf[13]) & 0x000000FF;
						l <<= 8;
						l |= hex2bin((char *)&rxLine.buf[15]) & 0x000000FF;
						if (c == active_sw_config) {
							configSw.fpga_start_address = l;
							// save the active configuration
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
							save_configuration(c, &configSw); // save the hardware + active software configurations
						} else {
							// non-active configuration. read from flash
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							cs.fpga_start_address = l;
							// save this non-active configuration
							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
							// Note: saving takes 120ms!! during which incoming messages are not processed
							save_configuration(c, &cs); // save the hardware + active software + this non-active configurations
						}
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "GCA", 3)) {
					// GCA - Get Configuration start Address command
					// read the selected configuration number for the specified start address
					// decode selected configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if (c < NSWCONFIGS) {
						if (c == active_sw_config) {
							// active configuration
							l = configSw.fpga_start_address;
						} else {
							// non-active configuration. read from flash
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							l = cs.fpga_start_address;
						}
						// Virtual location (not accounting bad-block redirection to another physical address)
						if (l == 0xFFFFFFFF) {
							// undefined
							snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CAS????????\r\n");
						} else {
							snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CAS%08X\r\n", l);
						}
						txLine.wptr = strlen((const char *)txLine.buf); // set message length
						sendLine(commandPortNo);
					}

				} else if (!memcmp((const char *)&rxLine.buf[4], "WFS", 3)) {
					// WFS - Write File Sector command
					// first check block number is valid
					// flash sectors 17 through 29 (32KB*13 = 416KB) are reserved for one FPGA configuration
					// Important: use the compress feature in Xilinx ISE to reduce file size.
					if ((sect_no < 17) || (sect_no > 29)) {
						retval = -1; // NAK
					} else {
						// make sure the block does not relate to a protected configuration
						c = block_to_config(sect_no);
						read_sw_config(c, &cs); // read the software configuration of interest from flash
						if ((c == 0) && (PROTECTION == 1)) {
							retval = -1; // boot configuration is protected
						} else if (cs.protection) {
							retval = -1; // configuration is protected
						} else {
							// disassemble the packet sequence number and data field
							packet_no = 0;
							for (i = 0; i < 4; i++)
								packet_no = packet_no * 10 + rxLine.buf[i + 10]
										- '0';
							// convert the data field from hex ASCII to binary. Up to 128 bytes in the data field.
							l = (rxLine.wptr - 16) / 2;
							j = 0;
							for (i = 14; i < rxLine.wptr - 2; i += 2) {
								rxLine.buf[j++] = hex2bin((char *)&rxLine.buf[i]); // save binary in place
							}

							if (packet_no == 1) {
								// first packet within a block. Reset the byte count.
								rx_byte_count = 0;
							}

							if (packet_no == 9999) {
								// last packet. Two cases: did we last end at a 256 byte boundary?
								if ((rx_byte_count & 0x000000FF) == 0) {
									// last write ended at a 256-byte boundary. No need for previous data in buf1 cache.
									// fill empty space with FFs
									memset((char *)&rxLine.buf[l], 0xFF, 256 - l);
								} else {
									// last write did not end at a 256-byte boundary.
									// Insert previous data in buf1 cache.
									memcpy((char *)&rxLine.buf[128], (const char *)&rxLine.buf[0],
											128); // make room for 128 bytes
									memcpy((char *)&rxLine.buf[0], (const char *)&buf1, 128); // insert previous 128 bytes
									// fill in buffer with FFs if needed
									if (l < 128) {
										memset((char *)&rxLine.buf[128 + l], 0xFF, 128
												- l);
									}
								}
								// write 256 Bytes (min number) to internal flash
								retval = internalFlash_write(sect_no,
										(rx_byte_count & 0xFFFFFF00),
										(char *)rxLine.buf, 256);
								rx_byte_count += l;
							} else if (((packet_no & 1) != 0) && (l == 128)) {
								// odd packet number (not last one). Can't write to internal flash until we gather 256 Bytes
								// Store until next WFS command
								memcpy(&buf1, (const char *)rxLine.buf, l);
								rx_byte_count += l;
								retval = 0; // OK
							} else if ((packet_no & 1) == 0) {
								// even packet number
								memcpy((char *)&rxLine.buf[128], (const char *)&rxLine.buf[0], 128); // make room for 128 bytes
								memcpy((char *)&rxLine.buf[0], (const char *)&buf1, 128); // insert previous 128 bytes
								// write size is limited to 256, 512, ... 4KB. No other values permitted.
								// fill in buffer with FFs if needed
								if (l < 128) {
									memset((char *)&rxLine.buf[128 + l], 0xFF, 128 - l);
								}
								// write 256 Bytes (min number) to internal flash
								retval = internalFlash_write(sect_no,
										(rx_byte_count & 0xFFFFFF00),
										(char *)rxLine.buf, 256);
								rx_byte_count += l;
							}
						}
					}

					if (retval) {
						// error -> NAK
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999ACK%06d\r\n", rx_byte_count);
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "GCS", 3)) {
					// GCS - Get CheckSum command
					retval = internalFlash_GCS(sect_no, (unsigned long *) (&c));

					// first check block number is valid
					if (retval) {
						// block number out-of-range error -> NAK
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999NAK\r\n");
					} else {
						// return the 16-bit checksum
						snprintf((char *)txLine.buf, sizeof(txLine.buf), "@999CKS%04X\r\n", c);
					}
					txLine.wptr = strlen((const char *)txLine.buf); // set message length
					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "PCF", 3)) {
					// PCF - Protect Configuration command
					// decode selected configuration number
					c = s78_decode - 1;
					// make sure configuration pointer is within allowed boundaries
					if (c < NSWCONFIGS) {
						if (c == active_sw_config) {
							if (rxLine.buf[9] == 'U') {
								// Unprotect configuration
								configSw.protection = 0;
							} else if (rxLine.buf[9] == 'P') {
								// Protect configuration
								configSw.protection = 1;
							}
							// save the active configuration
							configSw.save_request_pending = 1; // don't forget to save changes into non-volatile flash (cleared after saving)
							timer[1] = TIMER1SLOW;
							// re-arm timer[1]: wait before saving the comblock configuration
							//into flash (to consolidate multiple successive changes into one sector programming operation)
						} else {
							// non-active configuration. read from flash
							read_sw_config(c, &cs); // read the software configuration of interest from flash
							if (rxLine.buf[9] == 'U') {
								// Unprotect configuration
								cs.protection = 0;
							} else if (rxLine.buf[9] == 'P') {
								// Protect configuration
								cs.protection = 1;
							}
							// save this non-active configuration
							timer[1] = 0; // be safe: prevent another previously scheduled save from interfering.
							// Note: saving takes 120ms!! during which incoming messages are not processed
							save_configuration(c, &cs); // save the hardware + active software + this non-active configurations
						}
					}

//				} else if (!memcmp((const char *)&rxLine.buf[4], "TST", 3)) {
//					// test test test
//					// read data at sector 17
//					c = 0x00018000 + 128 * s789_decode;
//					memcpy(&buf1[0], (void *) (c), sizeof(buf1));
//					for (i = 0; i < sizeof(buf1); i++) {
//						snprintf((char *)&txLine.buf[2 * i], 3, "%02X", buf1[i]);
//					}
//					snprintf((char *)&txLine.buf[2 * i], 3, "\r\n");
//					txLine.wptr = strlen((const char *)txLine.buf); // set message length
//					sendLine(commandPortNo);

				} else if (!memcmp((const char *)&rxLine.buf[4], "RST", 3)) {
					// RST command - reset
					goto rst;
					/*
					 // TODO reset the FPGA
					 fpga_reset();
					 // reset micro
					 ResetISR();
					 */

				}

				// end of message processing section
				rxLine.rptr = rxLine.wptr; // release rx message


			}
		}
	}
	// end of main loop ------------------------------------------------------------------------


	return (0);
}

/*** EOF ***/
